/*
 * Copyright (c) 2021 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <net/if.h>

#include <linux/if.h>
#include <linux/if_link.h>
#include <linux/if_addr.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <libmnl/libmnl.h>

#include <avahi-common/thread-watch.h>
#include <avahi-common/error.h>
#include <avahi-client/client.h>
#include <avahi-client/publish.h>

char local_hostname[512] = "txt40.local";

static const char *aliases[] = {
    "txt40.local",
    NULL
};

static const char *ignored_interfaces[] = {
    "lo",
    NULL
};

static AvahiThreadedPoll *thread = NULL;
static AvahiClient *client = NULL;
static struct mnl_socket *mnl_sock = NULL;

struct address_entry;

struct alias_entry {
    struct alias_entry *next;
    struct address_entry *address;
    AvahiEntryGroup *group;
    const char *alias;
};

struct address_entry {
    struct address_entry *next;
    AvahiIfIndex index;
    AvahiAddress address;
    char *ip_addr;
    struct alias_entry *aliases;
};

static struct address_entry *addresses = NULL;

static void update_aliases()
{
   FILE *fptr;
   fptr = fopen("/etc/hostname", "r");
   if(fptr != NULL) 
   {
       fgets(local_hostname,sizeof(local_hostname),fptr);
       fclose(fptr);
       local_hostname[strlen(local_hostname)-1] = 0;
       strcat(local_hostname,".local");
   }
}

static void thread_quit()
{
    struct mnl_socket *s = mnl_sock;

    mnl_sock = NULL;
    if (s != NULL && mnl_socket_close(s) < 0) {
        perror("thread_quit: mnl_socket_close()");
    }
    if (thread != NULL && client != NULL) {
        /* First time client_cb is called in context of main process */
        avahi_threaded_poll_quit(thread);
    }
}

static void group_cb(AvahiEntryGroup *g, AvahiEntryGroupState state,
                     void *userdata)
{
    int ret;
    struct alias_entry *entry = (struct alias_entry *)userdata;

    switch (state) {
    case AVAHI_ENTRY_GROUP_ESTABLISHED:
        fprintf(stderr, "%s: Group Established: '%s' -> '%s' @ %d\n",
                __func__, entry->alias, entry->address->ip_addr,
                entry->address->index);
        break;

    case AVAHI_ENTRY_GROUP_COLLISION:
        fprintf(stderr, "%s: Group Collision: '%s' -> '%s' @ %d\n",
                __func__, entry->alias, entry->address->ip_addr,
                entry->address->index);
        ret = avahi_entry_group_reset(g);
        if (ret < 0) {
            fprintf(stderr, "%s: avahi_entry_group_reset() failed, error: %s\n",
                    __func__,
                    avahi_strerror(avahi_client_errno(
                                   avahi_entry_group_get_client(g))));
        }
        break;

    case AVAHI_ENTRY_GROUP_FAILURE:
        fprintf(stderr, "%s: Group Failure: '%s' -> '%s' @ %d, error: %s\n",
                __func__, entry->alias, entry->address->ip_addr,
                entry->address->index,
                avahi_strerror(avahi_client_errno(
                               avahi_entry_group_get_client(g))));
        /* thread_quit(); */
        break;

    case AVAHI_ENTRY_GROUP_UNCOMMITED:
        break;
    case AVAHI_ENTRY_GROUP_REGISTERING:
        break;
    }
}

static int register_hostnames(AvahiClient *c)
{
    const char **name;
    struct address_entry *address;
    struct alias_entry **alias;
    int ret;

    update_aliases();
    for (address = addresses; address != NULL; address = address->next) {
        for (alias = &address->aliases, name = aliases; *name != NULL;
             ++name, alias = &(*alias)->next) {

            if (*alias == NULL) {
                *alias = malloc(sizeof(struct alias_entry));
                if (*alias == NULL) {
                    perror("register_hostnames: malloc()");
                    return -1;
                }
                (*alias)->address = address;
                (*alias)->next = NULL;
                (*alias)->group = NULL;
            }
            (*alias)->alias = *name;

            if ((*alias)->group == NULL) {
                (*alias)->group = avahi_entry_group_new(c, group_cb, *alias);
                if ((*alias)->group == NULL) {
                    fprintf(stderr, "%s: avahi_entry_group_new() failed, "
                            "error: %s\n",
                            __func__, avahi_strerror(avahi_client_errno(c)));
                    return -1;
                }
            } else {
                ret = avahi_entry_group_reset((*alias)->group);
                if (ret < 0) {
                    fprintf(stderr, "%s: avahi_entry_group_reset() failed, "
                            "error: %s\n", __func__, avahi_strerror(ret));
                    return -1;
                }
            }

            ret = avahi_entry_group_add_address((*alias)->group, address->index,
                                                address->address.proto,
                                                AVAHI_PUBLISH_NO_REVERSE,
                                                *name, &address->address);
            if (ret < 0) {
                fprintf(stderr, "%s: avahi_entry_group_add_address() failed, "
                        " error: %s\n", __func__, avahi_strerror(ret));
                return -1;
            }

            ret = avahi_entry_group_commit((*alias)->group);
            if (ret < 0) {
                fprintf(stderr, "%s: avahi_entry_group_commit() failed, "
                        "error: %s\n", __func__, avahi_strerror(ret));
                return -1;
            }
            fprintf(stderr, "%s: Registered: '%s' -> '%s' @ %d\n",
                    __func__, *name, address->ip_addr, address->index);
        }
    }

    return 0;
}

static void client_cb(AvahiClient *c, AvahiClientState state,
                      AVAHI_GCC_UNUSED void *userdata)
{
    int ret;
    struct address_entry *address;
    struct alias_entry *alias;

    switch (state) {
    case AVAHI_CLIENT_FAILURE:
        if (avahi_client_errno(c) == AVAHI_ERR_DISCONNECTED) {
            fprintf(stderr, "%s: Disconnected, reconnecting...\n", __func__);
            update_aliases();
            for (address = addresses; address != NULL;
                                      address = address->next) {
                for (alias = address->aliases; alias != NULL;
                                               alias = alias->next) {
                    alias->group = NULL;
                }
            }
            avahi_client_free(c);
            client = avahi_client_new(avahi_threaded_poll_get(thread),
                                      AVAHI_CLIENT_NO_FAIL, client_cb,
                                      NULL, &ret);
            if (client == NULL) {
                fprintf(stderr, "%s: avahi_client_new() failed, error: %s\n",
                        __func__, avahi_strerror(ret));
                thread_quit();
            }
        } else {
            fprintf(stderr, "%s: Client failure: %s\n", __func__,
                    avahi_strerror(avahi_client_errno(c)));
            thread_quit();
        }
        break;

    case AVAHI_CLIENT_S_RUNNING:
        fprintf(stderr, "%s: Client running...\n", __func__);
        if (register_hostnames(c) < 0) {
            thread_quit();
        }
        break;

    case AVAHI_CLIENT_S_COLLISION:
        fprintf(stderr, "%s: Client collision\n", __func__);
        break;

    case AVAHI_CLIENT_S_REGISTERING:
        fprintf(stderr, "%s: Client registering...\n", __func__);
        /* avahi_entry_group_reset(group); */
        break;

    case AVAHI_CLIENT_CONNECTING:
        fprintf(stderr, "%s: Waiting for Avahi daemon...\n", __func__);
        break;
    }
}

static int address_attr_cb(const struct nlattr *attr, void *data)
{
    const struct nlattr **tb = data;
    int type = mnl_attr_get_type(attr);

    /* skip unsupported attribute in user-space */
    if (mnl_attr_type_valid(attr, IFA_MAX) < 0)
        return MNL_CB_OK;

    switch (type) {
    case IFA_ADDRESS:
    case IFA_LOCAL:
        if (mnl_attr_validate(attr, MNL_TYPE_BINARY) < 0) {
            perror("address_attr_cb: mnl_attr_validate()");
            return MNL_CB_ERROR;
        }
        break;
    }
    tb[type] = attr;
    return MNL_CB_OK;
}

static int address_cb(const struct nlmsghdr *msg, void *data)
{
    struct nlattr *tb[IFLA_MAX+1] = {};
    struct nlattr *tbaddr;
    struct ifaddrmsg *ifa = mnl_nlmsg_get_payload(msg);
    void *inaddr;
    char ip_addr[INET6_ADDRSTRLEN + 1];
    const char **name;
    int ret;

    if (msg->nlmsg_type != RTM_NEWADDR && msg->nlmsg_type != RTM_DELADDR)
        return MNL_CB_OK;

    if ((long)msg->nlmsg_len < (long)NLMSG_LENGTH(sizeof(*ifa))) {
        fprintf(stderr, "%s: Bad length: %ld < %ld\n", __func__,
                (long)msg->nlmsg_len, (long)NLMSG_LENGTH(sizeof(*ifa)));
        return MNL_CB_ERROR;
    }

    if (ifa->ifa_family != AF_INET && ifa->ifa_family != AF_INET6)
        return MNL_CB_OK;

    for (name = ignored_interfaces; *name != NULL; ++name) {
        if (ifa->ifa_index == if_nametoindex(*name))
            return MNL_CB_OK;
    }

    mnl_attr_parse(msg, sizeof(*ifa), address_attr_cb, tb);

    if (tb[IFA_LOCAL] == NULL && tb[IFA_ADDRESS] == NULL)
        return MNL_CB_OK;

    tbaddr = tb[IFA_LOCAL] != NULL ? tb[IFA_LOCAL] : tb[IFA_ADDRESS];
    inaddr = mnl_attr_get_payload(tbaddr);
    ip_addr[0] = '\0';
    if (inet_ntop(ifa->ifa_family, inaddr, ip_addr, sizeof(ip_addr)) == NULL) {
        perror("address_cb: inet_ntop()");
        return MNL_CB_ERROR;
    }
    ip_addr[INET6_ADDRSTRLEN] = '\0';

    if (msg->nlmsg_type == RTM_NEWADDR) {

        struct address_entry *address;

        fprintf(stderr, "%s: Added: index=%d family=%d scope=%d pfxlen=%d "
                "address=%s\n", __func__, ifa->ifa_index, ifa->ifa_family,
                ifa->ifa_scope, ifa->ifa_prefixlen, ip_addr);

        for (address = addresses; address != NULL; address = address->next) {
            if (strcasecmp(address->ip_addr, ip_addr) == 0) {
                break;
            }
        }

        if (address == NULL) {
            address = malloc(sizeof(struct address_entry));
            if (address == NULL) {
                perror("address_cb: malloc()");
                return MNL_CB_ERROR;
            }
            address->next = addresses;
            address->index = ifa->ifa_index;
            address->ip_addr = strdup(ip_addr);
            if (avahi_address_parse(ip_addr, avahi_af_to_proto(ifa->ifa_family),
                                    &address->address) == NULL) {
                perror("address_cb: avahi_address_parse() failed");
                return MNL_CB_ERROR;
            }
            address->aliases = NULL;
            addresses = address;
        }

        if (thread != NULL && client != NULL) {
            avahi_threaded_poll_lock(thread);

            if (avahi_client_get_state(client) == AVAHI_CLIENT_S_RUNNING) {
                if (register_hostnames(client) < 0) {
                    avahi_threaded_poll_unlock(thread);
                    return MNL_CB_ERROR;
                }
            }

            avahi_threaded_poll_unlock(thread);
        }

    } else if (msg->nlmsg_type == RTM_DELADDR) {

        struct address_entry **address;
        struct alias_entry *alias;
        void *a;

        fprintf(stderr, "%s: Deleted: index=%d family=%d scope=%d pfxlen=%d "
                "address=%s\n", __func__, ifa->ifa_index, ifa->ifa_family,
                ifa->ifa_scope, ifa->ifa_prefixlen, ip_addr);

        if (thread != NULL && client != NULL) {

            avahi_threaded_poll_lock(thread);

            for (address = &addresses; *address != NULL;) {
                if (strcasecmp((*address)->ip_addr, ip_addr) == 0
                        /* && (*address)->index == ifa->ifa_index */) {
                    update_aliases();
                    for (alias = (*address)->aliases; alias != NULL;) {
                        if (alias->group != NULL) {
                            ret = avahi_entry_group_reset(alias->group);
                            if (ret < 0) {
                                fprintf(stderr, "%s: avahi_entry_group_reset() "
                                        "failed, error: %s\n",
                                        __func__, avahi_strerror(ret));
                            }
                            ret = avahi_entry_group_free(alias->group);
                            if (ret < 0) {
                                fprintf(stderr, "%s: avahi_entry_group_free() "
                                        "failed, error: %s\n",
                                        __func__, avahi_strerror(ret));
                            }
                        }
                        a = alias;
                        alias = alias->next;
                        free(a);
                    }
                    free((*address)->ip_addr);
                    a = *address;
                    *address = (*address)->next;
                    free(a);
                } else {
                    address = &(*address)->next;
                }
            }

            avahi_threaded_poll_unlock(thread);
        }
    }
    return MNL_CB_OK;
}

int main(int argc, char **argv)
{
    struct nlmsghdr *msg;
    struct rtgenmsg *rtg;
    struct address_entry *address;
    struct alias_entry *alias;
    void *a;
    unsigned int sequence, portid;
    char buf[MNL_SOCKET_BUFFER_SIZE];
    int ret;

    /* Perform IP address enumeration */

    mnl_sock = mnl_socket_open(NETLINK_ROUTE);
    if (mnl_sock == NULL) {
        perror("main: mnl_socket_open()");
        return 1;
    }

    ret = mnl_socket_bind(mnl_sock, 0, MNL_SOCKET_AUTOPID);
    if (ret < 0) {
        perror("main: mnl_socket_bind()");
        goto cleanup;
    }

    msg = mnl_nlmsg_put_header(buf);
    msg->nlmsg_type = RTM_GETADDR;
    msg->nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
    msg->nlmsg_seq = time(NULL);
    rtg = mnl_nlmsg_put_extra_header(msg, sizeof(struct rtgenmsg));
    rtg->rtgen_family = PF_UNSPEC;

    portid = mnl_socket_get_portid(mnl_sock);
    sequence = msg->nlmsg_seq;

    ret = mnl_socket_sendto(mnl_sock, msg, msg->nlmsg_len);
    if (ret < 0) {
        perror("main: mnl_socket_sendto()");
        goto cleanup;
    }

    ret = mnl_socket_recvfrom(mnl_sock, buf, sizeof(buf));
    while (ret > 0) {
        ret = mnl_cb_run(buf, ret, sequence, portid, address_cb, NULL);
        if (ret <= MNL_CB_STOP)
            break;
        ret = mnl_socket_recvfrom(mnl_sock, buf, sizeof(buf));
    }
    if (ret < 0) {
        perror("main: mnl_socket_recvfrom()");
    }

    ret = mnl_socket_close(mnl_sock);
    if (ret < 0) {
        perror("main: mnl_socket_close()");
    }
    mnl_sock = NULL;

    /* Prepare Avahi client */

    thread = avahi_threaded_poll_new();
    if (thread == NULL) {
        fprintf(stderr, "%s: avahi_threaded_poll_new() failed\n", __func__);
        ret = -ENOMEM;
        goto cleanup;
    }

    client = avahi_client_new(avahi_threaded_poll_get(thread),
                              AVAHI_CLIENT_NO_FAIL, client_cb, NULL, &ret);
    if (client == NULL) {
        fprintf(stderr, "%s: avahi_client_new() failed, error: %s\n",
                __func__, avahi_strerror(ret));
        goto cleanup;
    }

    /* Prepare IP address event listening socket */

    mnl_sock = mnl_socket_open(NETLINK_ROUTE);
    if (mnl_sock == NULL) {
        perror("main: mnl_socket_open()");
        ret = -ENOMEM;
        goto cleanup;
    }

    ret = mnl_socket_bind(mnl_sock, RTMGRP_IPV4_IFADDR | RTMGRP_IPV6_IFADDR,
                          MNL_SOCKET_AUTOPID);
    if (ret < 0) {
        perror("main: mnl_socket_bind()");
        goto cleanup;
    }

    /* Now start Avahi client */

    ret = avahi_threaded_poll_start(thread);
    if (ret < 0) {
        fprintf(stderr, "%s: avahi_threaded_poll_start() failed, error: %s\n",
                __func__, avahi_strerror(ret));
        goto cleanup;
    }

    ret = mnl_socket_recvfrom(mnl_sock, buf, sizeof(buf));
    while (ret > 0) {
        ret = mnl_cb_run(buf, ret, 0, 0, address_cb, NULL);
        if (ret <= MNL_CB_STOP)
            break;
        ret = mnl_socket_recvfrom(mnl_sock, buf, sizeof(buf));
    }
    if (ret < 0) {
        perror("main: mnl_socket_recvfrom()");
    }

    ret = avahi_threaded_poll_stop(thread);
    if (ret < 0) {
        fprintf(stderr, "%s: avahi_threaded_poll_stop() failed, error: %s\n",
                __func__, avahi_strerror(ret));
    }

    ret = 0;

cleanup:
    if (mnl_sock != NULL) {
        ret = mnl_socket_close(mnl_sock);
        if (ret < 0) {
            perror("main: mnl_socket_close()");
        }
    }

    if (client != NULL) {
        avahi_client_free(client);
    }

    if (thread != NULL) {
        avahi_threaded_poll_free(thread);
    }

    for (address = addresses; address != NULL;) {
        update_aliases();
        for (alias = address->aliases; alias != NULL;) {
            a = alias;
            alias = alias->next;
            free(a);
        }
        free(address->ip_addr);
        a = address;
        address = address->next;
        free(a);
    }

    return ret;
}
